using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.Collections;
using System.IO;
using System.Globalization;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters;
using System.Runtime.Serialization.Formatters.Binary;

using ProjectCarbon.Properties;
using ProjectCarbon.Protocols;

namespace ProjectCarbon
{
    public partial class frmMain : Form
    {
        public frmMain()
        {
            InitializeComponent();
        }

        private void frmMain_Load(object sender, EventArgs e)
        {
            StringBuilder ipAddressesString = new StringBuilder();

            if (Global.LocalIPAddresses.Length == 0)
            {
                throw new Exception("No IP addresses found on this machine!");
            }
            else if (Global.LocalIPAddresses.Length == 1)
            {
                ipAddressesStatus.Text = String.Format(CultureInfo.CurrentCulture, 
                    "IP Address: {0}", Global.LocalIPAddresses[0].ToString());
            }
            else
            {
                ipAddressesString.Append("IP Addresses: ");
                for (int i = 0; i < Global.LocalIPAddresses.Length; i++)
                    if (i == (Global.LocalIPAddresses.Length - 1))
                        ipAddressesString.Append(Global.LocalIPAddresses[i].ToString());
                    else
                        ipAddressesString.Append(Global.LocalIPAddresses[i].ToString() + ", ");
                ipAddressesStatus.Text = ipAddressesString.ToString();
            }

            Global.PacketCaptureArray = new PacketCapture[Global.LocalIPAddresses.Length];
            for (int i = 0; i < Global.LocalIPAddresses.Length; i++)
            {
                Global.PacketCaptureArray[i] = new PacketCapture(Global.LocalIPAddresses[i]);
                Global.PacketCaptureArray[i].OnPacketCaptured += 
                    new PacketCapture.PacketCapturedEventHandler(frmMain_OnPacketCaptured);
            }

            spynetStatusPanel.Text = "Status: Idle...";
            cboDisplayModes.SelectedIndex = 0;

            // Let's load the options
            richTextPacketDisplay.BackColor =
                Settings.Default.PacketDisplayBackgroundColor;
            richTextPacketDisplay.ForeColor =
                Settings.Default.PacketDisplayFontColor;
            richTextPacketDisplay.Font =
                Settings.Default.PacketDisplayFontFace;
            packetCaptureList.BackColor =
                Settings.Default.PacketListBackgroundColor;
            packetCaptureList.ForeColor =
                Settings.Default.PacketListTextColor;

            // This event is fired when the properties are saving
            Settings.Default.SettingsSaving += 
                new System.Configuration.SettingsSavingEventHandler(Default_SettingsSaving);
        }


        void Default_SettingsSaving(object sender, CancelEventArgs e)
        {
            richTextPacketDisplay.BackColor =
                ProjectCarbon.Properties.Settings.Default.PacketDisplayBackgroundColor;
            richTextPacketDisplay.ForeColor =
                ProjectCarbon.Properties.Settings.Default.PacketDisplayFontColor;
            richTextPacketDisplay.Font =
                ProjectCarbon.Properties.Settings.Default.PacketDisplayFontFace;
            packetCaptureList.BackColor =
                ProjectCarbon.Properties.Settings.Default.PacketListBackgroundColor;
            packetCaptureList.ForeColor =
                ProjectCarbon.Properties.Settings.Default.PacketListTextColor;

            if (ProjectCarbon.Properties.Settings.Default.StartUpWindows)
                Functions.SetStartup();
            else
                Functions.RemoveStartup();
            
        }

        void frmMain_OnPacketCaptured(object sender, PacketCapturedEventArgs e)
        {
            IpProtocol iphp = new IpProtocol(e.Buffer);

            Global.PacketCapturedEvent pke
                 = new Global.PacketCapturedEvent(PacketCaptured);

            this.Invoke(pke, new object[] { iphp });
        }

        private void PacketCaptured(IpProtocol iphp)
        {
            IpSubProtocol packetProcessor =
                IpSubProtocol.GetProtocolInstance((Protocol)iphp.Properties["Protocol"]);
            
            packetProcessor.ProcessData((byte[])iphp.Properties["Payload"]);

            bool portFilter = Global.PortFilterInclusive;
            bool protocolFilter = Global.ProtocolFilterInclusive;

            if (Global.FilteredPortList.Count > 0)
            {
                if (packetProcessor.Properties.ContainsKey("SourcePort") == true &&
                    packetProcessor.Properties.ContainsKey("DestinationPort") == true)
                {
                    for (int i = 0; i < Global.FilteredPortList.Count; i++)
                    {
                        if (Global.FilteredPortList[i] == (int)packetProcessor.Properties["SourcePort"] ||
                            Global.FilteredPortList[i] == (int)packetProcessor.Properties["DestinationPort"])
                        {
                            portFilter = (Global.PortFilterInclusive ? false : true);
                            break;
                        }
                    }
                }
            }
            else
            {
                portFilter = false;
            }

            if (Global.FilteredProtocolList.Count > 0)
            {
                for (int i = 0; i < Global.FilteredProtocolList.Count; i++)
                {
                    if (Global.FilteredProtocolList[i].ToLower() ==
                        iphp.Properties["Protocol"].ToString().ToLower())
                    {
                        protocolFilter = (Global.ProtocolFilterInclusive ? false : true);
                        break;
                    }
                }
            }
            else
                protocolFilter = false;

            // Blocked
            if (portFilter == true || protocolFilter == true) 
                return;

            ListViewItem lvi = packetCaptureList.Items.Add(packetCaptureList.Items.Count.ToString());

            lvi.SubItems.Add(((DateTime)(iphp.Properties["PacketTime"])).ToString("HH:mm:ss"));
            lvi.SubItems.Add(iphp.Properties["Protocol"].ToString());
            lvi.SubItems.Add((string)iphp.Properties["SourceAddress"]);
            lvi.SubItems.Add((string)iphp.Properties["DestinationAddress"]);

            if (packetProcessor.Properties.ContainsKey("SourcePort") == true &&
                packetProcessor.Properties.ContainsKey("DestinationPort") == true)
            {
                lvi.SubItems.Add(packetProcessor.Properties["SourcePort"].ToString());
                lvi.SubItems.Add(packetProcessor.Properties["DestinationPort"].ToString());
            } else {
                lvi.SubItems.Add("--");
                lvi.SubItems.Add("--");
            }

            // Add this captured packet to the list.
            Global.CapturedPackets.Add(iphp);

            // Update the packet list.
            packetsInBufferStatus.Text = "Captured Packets: " + Global.CapturedPackets.Count;

            // Scroll if autoscroll is enabled.
            if (Settings.Default.AutoScrollPacketList == true)
            {
                //lvi.Selected = true;
                lvi.EnsureVisible();
            }
        }

        private void DisplayCapture(byte[] buffer)
        {
            richTextPacketDisplay.Clear();
            switch (cboDisplayModes.SelectedIndex)
            {
                case 0:
                    richTextPacketDisplay.AppendText(Functions.GetHexBlock(buffer));
                    break;
                case 1:
                    richTextPacketDisplay.AppendText(Functions.GetHexOnly(buffer));
                    break;
                case 2:
                    richTextPacketDisplay.AppendText(Functions.GetDecBlock(buffer));
                    break;
                case 3:
                    richTextPacketDisplay.AppendText(Functions.GetDecOnly(buffer));
                    break;
                case 4:
                    richTextPacketDisplay.AppendText(Functions.GetSafeString(buffer));
                    break;
            }
        }

        private void btnStartCapture_Click(object sender, EventArgs e)
        {
            StartCapture();
        }

        private void packetCaptureList_ItemSelectionChanged(object sender, ListViewItemSelectionChangedEventArgs e)
        {
            IpProtocol iphp = (IpProtocol)Global.CapturedPackets[e.ItemIndex];
            IpSubProtocol packetProcessor =
                IpSubProtocol.GetProtocolInstance((Protocol)iphp.Properties["Protocol"]);

            packetProcessor.ProcessData((byte[])iphp.Properties["Payload"]);


            // Delete all nodes
            tvwPacketHeaderInfo.Nodes.Clear();

            tvwPacketHeaderInfo.BeginUpdate();

            // Display IP Header
            tvwPacketHeaderInfo.Nodes.Add("ip", "IP Header");
            foreach (KeyValuePair<string, object> kvpProp in iphp.Properties)
            {
                if (kvpProp.Key != "Payload" && 
                    kvpProp.Key != "PacketPayload")
                {
                    tvwPacketHeaderInfo.Nodes[0].Nodes.Add(kvpProp.Key + ": " +
                        kvpProp.Value.ToString());
                }
            }

            // Get's the protocols name and properties
            tvwPacketHeaderInfo.Nodes.Add(packetProcessor.Name, 
                packetProcessor.Name + " Header");
            // Cycle through the different properties
            foreach (KeyValuePair<string, object> kvpProp in packetProcessor.Properties)
            {
                if (kvpProp.Key != "Payload")
                {
                    tvwPacketHeaderInfo.Nodes[1].Nodes.Add(kvpProp.Key + ": " +
                        kvpProp.Value.ToString());
                }
            }

            //TODO: Add a system to process out subprotocols within the IP protocols.
            
            // See, I designed this so that if a protocol does have a payload
            // then it will be in the property "Payload".
            if (chkStripTCPIPHeaders.Checked == true)
                DisplayCapture((byte[])packetProcessor.Properties["Payload"]);
            else
                DisplayCapture((byte[])iphp.Properties["PacketPayload"]);

            // I expand these backwards, otherwise it will scroll to the bottom.
            tvwPacketHeaderInfo.ExpandAll();
            tvwPacketHeaderInfo.Nodes[0].EnsureVisible();

            tvwPacketHeaderInfo.EndUpdate();

        }

        private void btnStopCapture_Click(object sender, EventArgs e)
        {
            StopCapture();
        }

        private void mnuAbout_Click(object sender, EventArgs e)
        {
            frmAboutDialog aboutDialog = new frmAboutDialog();
            aboutDialog.ShowDialog();
        }

        private void cboDisplayModes_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (packetCaptureList.SelectedItems.Count > 0)
                packetCaptureList_ItemSelectionChanged(packetCaptureList, new ListViewItemSelectionChangedEventArgs(packetCaptureList.SelectedItems[0], packetCaptureList.SelectedItems[0].Index, true));
        }

        private void btnClearCaptureLog_Click(object sender, EventArgs e)
        {
            ClearCaptureLog();
        }

        private void btnSaveCaptureFile_Click(object sender, EventArgs e)
        {
            SaveCaptureFile();
        }

        private void btnLoadCaptureFile_Click(object sender, EventArgs e)
        {
            LoadCaptureFile(false);
        }

        private void btnCaptureFilters_Click(object sender, EventArgs e)
        {
            ShowCaptureFiltersDialog();
        }

        private void chkStripTCPIPHeaders_CheckStateChanged(object sender, EventArgs e)
        {
            if (packetCaptureList.SelectedItems.Count > 0)
                packetCaptureList_ItemSelectionChanged(packetCaptureList, new ListViewItemSelectionChangedEventArgs(packetCaptureList.SelectedItems[0], packetCaptureList.SelectedItems[0].Index, true));
        }

        private void mnuOpenCaptureFile_Click(object sender, EventArgs e)
        {
            LoadCaptureFile(false);
        }

        private void mnuOpenAndAppend_Click(object sender, EventArgs e)
        {
            LoadCaptureFile(true);
        }

        private void mnuStartCapture_Click(object sender, EventArgs e)
        {
            StartCapture();
        }

        private void mnuStopCapture_Click(object sender, EventArgs e)
        {
            StopCapture();
        }

        private void mnuClearCapturedPackets_Click(object sender, EventArgs e)
        {
            ClearCaptureLog();
        }

        private void mnuSettings_Click(object sender, EventArgs e)
        {
            ShowCaptureSettings();
        }

        private void mnuFilters_Click(object sender, EventArgs e)
        {
            ShowCaptureFiltersDialog();
        }

        private void mnuSaveCaptureFile_Click(object sender, EventArgs e)
        {
            SaveCaptureFile();
        }

        private void btnCaptureSettings_Click(object sender, EventArgs e)
        {
            ShowCaptureSettings();
        }


        private void mnuDisplayWinsockInfo_Click(object sender, EventArgs e)
        {
            frmSocketInformation socketInformation = new frmSocketInformation();
            socketInformation.ShowDialog();
        }

        private void mnuAdapterPerformance_Click(object sender, EventArgs e)
        {
            frmAdapterPerformance adapterPerformance = new frmAdapterPerformance();
            adapterPerformance.ShowDialog();
        }

        private void mnuVisitWebsite_Click(object sender, EventArgs e)
        {
            Help.ShowHelp(this, "http://www.xeons.net/");
        }

        private void frmMain_Resize(object sender, EventArgs e)
        {
            if (FormWindowState.Minimized == WindowState)
                Hide();
        }

        private void niTrayIcon_DoubleClick(object sender, EventArgs e)
        {
            Show();
            WindowState = FormWindowState.Normal;
        }

        private void frmMain_Shown(object sender, EventArgs e)
        {
            // Start capturing at startup
            if (Settings.Default.StartUpCapture)
                StartCapture();

            // Startup Minimized to tray
            if (Settings.Default.StartUpMinimized)
                WindowState = FormWindowState.Minimized;
        }

        private void savePacketDumpToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SavePacketDump();
        }

        private void mnuTrayStartCapture_Click(object sender, EventArgs e)
        {
            StartCapture();
        }

        private void mnuTrayStopCapture_Click(object sender, EventArgs e)
        {
            StopCapture();
        }

        private void mnuTraySettings_Click(object sender, EventArgs e)
        {
            ShowCaptureSettings();
        }

        private void mnuTrayFilters_Click(object sender, EventArgs e)
        {
            ShowCaptureFiltersDialog();
        }

        private void StartCapture()
        {
            for (int i = 0; i < Global.LocalIPAddresses.Length; i++)
            {
                Global.PacketCaptureArray[i].StartCapture();
            }
            spynetStatusPanel.Text = "Status: Capturing Packets...";
        }

        private void StopCapture()
        {
            for (int i = 0; i < Global.LocalIPAddresses.Length; i++)
            {
                Global.PacketCaptureArray[i].StopCapture();
            }
            spynetStatusPanel.Text = "Status: Capture Stopped.";
        }

        private void ClearCaptureLog()
        {
            if (MessageBox.Show("Are you sure you want to clear the captured packets list?", 
                "Project Carbon", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes)
            {
                packetCaptureList.Items.Clear();
                Global.CapturedPackets.Clear();
                packetsInBufferStatus.Text = "Captured Packets: " + Global.CapturedPackets.Count;
            }
        }

        private void LoadCaptureFile(bool append)
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();
            
            openFileDialog.CheckFileExists = true;
            openFileDialog.CheckPathExists = true;
            openFileDialog.Title = "Load Capture";
            openFileDialog.Filter = "ProjectCarbon Capture File (*.cap2)|*.cap2|CaptureNet Files (*.cap)|*.cap|All Files (*.*)|*.*";
            
            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                BinaryFormatter bf = new BinaryFormatter();
                // Let's open the capture file and start reading it.
                using (FileStream fs = new FileStream(openFileDialog.FileName, 
                    FileMode.Open))
                {
                    BinaryReader br = new BinaryReader(fs);

                    byte nameLength = br.ReadByte();
                    byte[] fileIdentifier = new byte[5];

                    if (nameLength > 1 && nameLength < 20)
                        fileIdentifier = br.ReadBytes(nameLength);

                    if (Encoding.ASCII.GetString(fileIdentifier).IndexOf("SpyNet") != -1)
                    {
                        do
                        {
                            ushort packetLength = br.ReadUInt16();
                            byte[] preData = br.ReadBytes(28);
                            byte[] packetData = br.ReadBytes(packetLength - 14);
                            if (packetData.Length > 0)
                            {
                                PacketCaptured(new IpProtocol(packetData));
                            }
                        }
                        while (br.BaseStream.Position < br.BaseStream.Length);
                        return;
                    }

                    fs.Position = 0;

                    try
                    {
                        if (!append)
                        {
                            packetCaptureList.Items.Clear();
                            Global.CapturedPackets.Clear();
                        }

                        List<IpProtocol> capturedPacketsLoaded = (List<IpProtocol>)bf.Deserialize(fs);
                        foreach (IpProtocol ipp in capturedPacketsLoaded.ToArray())
                        {
                            PacketCaptured(ipp);
                        }
                    }
                    catch (FileLoadException)
                    {
                    }
                    catch (SerializationException)
                    {
                        MessageBox.Show("There was an error while attempting to load this capture file. It may have become corrupted or obsolete.",
                            "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                }
            }
        }

        private static void SaveCaptureFile()
        {
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            
            saveFileDialog.CheckFileExists = false;
            saveFileDialog.CheckPathExists = true;
            saveFileDialog.Title = "Save Capture";
            saveFileDialog.Filter = "ProjectCarbon Capture File (*.cap2)|*.cap2|All Files|*.*";
            
            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                BinaryFormatter bf = new BinaryFormatter();
                using (FileStream fs = new FileStream(saveFileDialog.FileName,
                    FileMode.OpenOrCreate))
                {
                    bf.Serialize(fs, Global.CapturedPackets);
                }
            }
        }

        private static void SavePacketDump()
        {
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            int i = 0;
            IpSubProtocol packetProcessor = null;

            saveFileDialog.CheckFileExists = false;
            saveFileDialog.CheckPathExists = true;
            saveFileDialog.Title = "Save Capture";
            saveFileDialog.Filter = "Packet Dump (*.txt)|*.txt|All Files|*.*";

            if (Global.CapturedPackets.Count == 0)
            {
                MessageBox.Show("There is no captured packets to save.", "Error", 
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                using (TextWriter tw = new StreamWriter(saveFileDialog.FileName))
                {
                    foreach (IpProtocol iphp in Global.CapturedPackets)
                    {
                        packetProcessor =
                            IpSubProtocol.GetProtocolInstance((Protocol)iphp.Properties["Protocol"]);

                        packetProcessor.ProcessData((byte[])iphp.Properties["Payload"]);

                        tw.WriteLine("Packet Index: {0}", i++);
                        tw.WriteLine();

                        tw.WriteLine("IP Header:");
                        foreach (KeyValuePair<string, object> kvpProp in iphp.Properties)
                        {
                            if (kvpProp.Key != "Payload" &&
                                kvpProp.Key != "PacketPayload")
                            {
                                tw.WriteLine("{0}: {1}", kvpProp.Key, 
                                    kvpProp.Value.ToString());
                            }
                        }
                        tw.WriteLine();

                        tw.WriteLine("{0} Header:", packetProcessor.Name);
                        foreach (KeyValuePair<string, object> kvpProp in packetProcessor.Properties)
                        {
                            if (kvpProp.Key != "Payload" && 
                                kvpProp.Key != "PacketPayload")
                            {
                                tw.WriteLine("{0}: {1}", kvpProp.Key, 
                                    kvpProp.Value.ToString());
                            }
                        }
                        
                        tw.WriteLine();

                        tw.WriteLine("Packet data:");
                        tw.Write(Functions.GetHexBlock( (byte[])iphp.Properties["Payload"]) );
                        tw.WriteLine();
                        tw.WriteLine();
                        tw.WriteLine("=====================================================================");
                        tw.WriteLine();
                    }
                }
            }
        }

        private static void ShowCaptureFiltersDialog()
        {
            frmCaptureFilters captureFilters = new frmCaptureFilters();
            captureFilters.ShowDialog();
        }

        private static void ShowCaptureSettings()
        {
            frmCaptureSettings captureSettings = new frmCaptureSettings();
            captureSettings.ShowDialog();
        }

        private void mnuExit_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void mnuDnsLookup_Click(object sender, EventArgs e)
        {
            frmDnsLookup dnsLookup = new frmDnsLookup();
            dnsLookup.ShowDialog();
        }

        private void chkStripTCPIPHeaders_CheckedChanged(object sender, EventArgs e)
        {

        }
    }
}